# Copyright (c) 2015-2019, Activision Publishing, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sys

if sys.version_info[0] == 3:
    str_types = (str,)
    xrange = range
else:
    str_types = (basestring,)
    xrange = xrange

__tracebackhide__ = True


class ContainsMixin(object):
    """Containment assertions mixin."""

    def contains(self, *items):
        """Asserts that val contains the given item or items.

        Checks if the collection contains the given item or items using ``in`` operator.

        Args:
            *items: the item or items expected to be contained

        Examples:
            Usage::

                assert_that('foo').contains('f')
                assert_that('foo').contains('f', 'oo')
                assert_that(['a', 'b']).contains('b', 'a')
                assert_that((1, 2, 3)).contains(3, 2, 1)
                assert_that({'a': 1, 'b': 2}).contains('b', 'a')  # checks keys
                assert_that({'a', 'b'}).contains('b', 'a')
                assert_that([1, 2, 3]).is_type_of(list).contains(1, 2).does_not_contain(4, 5)

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val does **not** contain the item or items

        Tip:
            Use the :meth:`~assertpy.dict.DictMixin.contains_key` alias when working with
            *dict-like* objects to be self-documenting.

        See Also:
            :meth:`~assertpy.string.StringMixin.contains_ignoring_case` - for case-insensitive string contains
        """
        if len(items) == 0:
            raise ValueError('one or more args must be given')
        elif len(items) == 1:
            if items[0] not in self.val:
                if self._check_dict_like(self.val, return_as_bool=True):
                    self.error('Expected <%s> to contain key <%s>, but did not.' % (self.val, items[0]))
                else:
                    self.error('Expected <%s> to contain item <%s>, but did not.' % (self.val, items[0]))
        else:
            missing = []
            for i in items:
                if i not in self.val:
                    missing.append(i)
            if missing:
                if self._check_dict_like(self.val, return_as_bool=True):
                    self.error('Expected <%s> to contain keys %s, but did not contain key%s %s.' % (
                        self.val, self._fmt_items(items), '' if len(missing) == 0 else 's', self._fmt_items(missing)))
                else:
                    self.error('Expected <%s> to contain items %s, but did not contain %s.' % (self.val, self._fmt_items(items), self._fmt_items(missing)))
        return self

    def does_not_contain(self, *items):
        """Asserts that val does not contain the given item or items.

        Checks if the collection excludes the given item or items using ``in`` operator.

        Args:
            *items: the item or items expected to be excluded

        Examples:
            Usage::

                assert_that('foo').does_not_contain('x')
                assert_that(['a', 'b']).does_not_contain('x', 'y')
                assert_that((1, 2, 3)).does_not_contain(4, 5)
                assert_that({'a': 1, 'b': 2}).does_not_contain('x', 'y')  # checks keys
                assert_that({'a', 'b'}).does_not_contain('x', 'y')
                assert_that([1, 2, 3]).is_type_of(list).contains(1, 2).does_not_contain(4, 5)

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val **does** contain the item or items

        Tip:
            Use the :meth:`~assertpy.dict.DictMixin.does_not_contain_key` alias when working with
            *dict-like* objects to be self-documenting.
        """
        if len(items) == 0:
            raise ValueError('one or more args must be given')
        elif len(items) == 1:
            if items[0] in self.val:
                self.error('Expected <%s> to not contain item <%s>, but did.' % (self.val, items[0]))
        else:
            found = []
            for i in items:
                if i in self.val:
                    found.append(i)
            if found:
                self.error('Expected <%s> to not contain items %s, but did contain %s.' % (self.val, self._fmt_items(items), self._fmt_items(found)))
        return self

    def contains_only(self, *items):
        """Asserts that val contains *only* the given item or items.

        Checks if the collection contains only the given item or items using ``in`` operator.

        Args:
            *items: the *only* item or items expected to be contained

        Examples:
            Usage::

                assert_that('foo').contains_only('f', 'o')
                assert_that(['a', 'a', 'b']).contains_only('a', 'b')
                assert_that((1, 1, 2)).contains_only(1, 2)
                assert_that({'a': 1, 'a': 2, 'b': 3}).contains_only('a', 'b')
                assert_that({'a', 'a', 'b'}).contains_only('a', 'b')

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val contains anything **not** item or items
        """
        if len(items) == 0:
            raise ValueError('one or more args must be given')
        else:
            extra = []
            for i in self.val:
                if i not in items:
                    extra.append(i)
            if extra:
                self.error('Expected <%s> to contain only %s, but did contain %s.' % (self.val, self._fmt_items(items), self._fmt_items(extra)))

            missing = []
            for i in items:
                if i not in self.val:
                    missing.append(i)
            if missing:
                self.error('Expected <%s> to contain only %s, but did not contain %s.' % (self.val, self._fmt_items(items), self._fmt_items(missing)))
        return self

    def contains_sequence(self, *items):
        """Asserts that val contains the given ordered sequence of items.

        Checks if the collection contains the given sequence of items using ``in`` operator.

        Args:
            *items: the sequence of items expected to be contained

        Examples:
            Usage::

                assert_that('foo').contains_sequence('f', 'o')
                assert_that('foo').contains_sequence('o', 'o')
                assert_that(['a', 'b', 'c']).contains_sequence('b', 'c')
                assert_that((1, 2, 3)).contains_sequence(1, 2)

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val does **not** contains the given sequence of items
        """
        if len(items) == 0:
            raise ValueError('one or more args must be given')
        else:
            try:
                for i in xrange(len(self.val) - len(items) + 1):
                    for j in xrange(len(items)):
                        if self.val[i+j] != items[j]:
                            break
                    else:
                        return self
            except TypeError:
                raise TypeError('val is not iterable')
        self.error('Expected <%s> to contain sequence %s, but did not.' % (self.val, self._fmt_items(items)))

    def contains_duplicates(self):
        """Asserts that val is iterable and *does* contain duplicates.

        Examples:
            Usage::

                assert_that('foo').contains_duplicates()
                assert_that(['a', 'a', 'b']).contains_duplicates()
                assert_that((1, 1, 2)).contains_duplicates()

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val does **not** contain any duplicates
        """
        try:
            if len(self.val) != len(set(self.val)):
                return self
        except TypeError:
            raise TypeError('val is not iterable')
        self.error('Expected <%s> to contain duplicates, but did not.' % self.val)

    def does_not_contain_duplicates(self):
        """Asserts that val is iterable and *does not* contain any duplicates.

        Examples:
            Usage::

                assert_that('fox').does_not_contain_duplicates()
                assert_that(['a', 'b', 'c']).does_not_contain_duplicates()
                assert_that((1, 2, 3)).does_not_contain_duplicates()

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val **does** contain duplicates
        """
        try:
            if len(self.val) == len(set(self.val)):
                return self
        except TypeError:
            raise TypeError('val is not iterable')
        self.error('Expected <%s> to not contain duplicates, but did.' % self.val)

    def is_empty(self):
        """Asserts that val is empty.

        Examples:
            Usage::

                assert_that('').is_empty()
                assert_that([]).is_empty()
                assert_that(()).is_empty()
                assert_that({}).is_empty()
                assert_that(set()).is_empty()

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val is **not** empty
        """
        if len(self.val) != 0:
            if isinstance(self.val, str_types):
                self.error('Expected <%s> to be empty string, but was not.' % self.val)
            else:
                self.error('Expected <%s> to be empty, but was not.' % self.val)
        return self

    def is_not_empty(self):
        """Asserts that val is *not* empty.

        Examples:
            Usage::

                assert_that('foo').is_not_empty()
                assert_that(['a', 'b']).is_not_empty()
                assert_that((1, 2, 3)).is_not_empty()
                assert_that({'a': 1, 'b': 2}).is_not_empty()
                assert_that({'a', 'b'}).is_not_empty()

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val **is** empty
        """
        if len(self.val) == 0:
            if isinstance(self.val, str_types):
                self.error('Expected not empty string, but was empty.')
            else:
                self.error('Expected not empty, but was empty.')
        return self

    def is_in(self, *items):
        """Asserts that val is equal to one of the given items.

        Args:
            *items: the items expected to contain val

        Examples:
            Usage::

                assert_that('foo').is_in('foo', 'bar', 'baz')
                assert_that(1).is_in(0, 1, 2, 3)

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val is **not** in the given items
        """
        if len(items) == 0:
            raise ValueError('one or more args must be given')
        else:
            for i in items:
                if self.val == i:
                    return self
        self.error('Expected <%s> to be in %s, but was not.' % (self.val, self._fmt_items(items)))

    def is_not_in(self, *items):
        """Asserts that val is not equal to one of the given items.

        Args:
            *items: the items expected to exclude val

        Examples:
            Usage::

                assert_that('foo').is_not_in('bar', 'baz', 'box')
                assert_that(1).is_not_in(-1, -2, -3)

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val **is** in the given items
        """
        if len(items) == 0:
            raise ValueError('one or more args must be given')
        else:
            for i in items:
                if self.val == i:
                    self.error('Expected <%s> to not be in %s, but was.' % (self.val, self._fmt_items(items)))
        return self



# Copyright (c) 2015-2019, Activision Publishing, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sys
import numbers
import datetime
import collections

if sys.version_info[0] == 3:
    Iterable = collections.abc.Iterable
else:
    Iterable = collections.Iterable

__tracebackhide__ = True


class HelpersMixin(object):
    """Helpers mixin.  For internal use only."""

    def _fmt_items(self, i):
        """Helper to format the given items."""
        if len(i) == 0:
            return '<>'
        elif len(i) == 1 and hasattr(i, '__getitem__'):
            return '<%s>' % i[0]
        else:
            return '<%s>' % str(i).lstrip('([').rstrip(',])')

    def _fmt_args_kwargs(self, *some_args, **some_kwargs):
        """Helper to convert the given args and kwargs into a string."""
        if some_args:
            out_args = str(some_args).lstrip('(').rstrip(',)')
        if some_kwargs:
            out_kwargs = ', '.join([str(i).lstrip('(').rstrip(')').replace(', ', ': ') for i in [
                    (k, some_kwargs[k]) for k in sorted(some_kwargs.keys())]])

        if some_args and some_kwargs:
            return out_args + ', ' + out_kwargs
        elif some_args:
            return out_args
        elif some_kwargs:
            return out_kwargs
        else:
            return ''

    def _validate_between_args(self, val_type, low, high):
        """Helper to validate given range args."""
        low_type = type(low)
        high_type = type(high)

        if val_type in self._NUMERIC_NON_COMPAREABLE:
            raise TypeError('ordering is not defined for type <%s>' % val_type.__name__)

        if val_type in self._NUMERIC_COMPAREABLE:
            if low_type is not val_type:
                raise TypeError('given low arg must be <%s>, but was <%s>' % (val_type.__name__, low_type.__name__))
            if high_type is not val_type:
                raise TypeError('given high arg must be <%s>, but was <%s>' % (val_type.__name__, low_type.__name__))
        elif isinstance(self.val, numbers.Number):
            if isinstance(low, numbers.Number) is False:
                raise TypeError('given low arg must be numeric, but was <%s>' % low_type.__name__)
            if isinstance(high, numbers.Number) is False:
                raise TypeError('given high arg must be numeric, but was <%s>' % high_type.__name__)
        else:
            raise TypeError('ordering is not defined for type <%s>' % val_type.__name__)

        if low > high:
            raise ValueError('given low arg must be less than given high arg')

    def _validate_close_to_args(self, val, other, tolerance):
        """Helper for validate given arg and delta."""
        if type(val) is complex or type(other) is complex or type(tolerance) is complex:
            raise TypeError('ordering is not defined for complex numbers')

        if isinstance(val, numbers.Number) is False and type(val) is not datetime.datetime:
            raise TypeError('val is not numeric or datetime')

        if type(val) is datetime.datetime:
            if type(other) is not datetime.datetime:
                raise TypeError('given arg must be datetime, but was <%s>' % type(other).__name__)
            if type(tolerance) is not datetime.timedelta:
                raise TypeError('given tolerance arg must be timedelta, but was <%s>' % type(tolerance).__name__)
        else:
            if isinstance(other, numbers.Number) is False:
                raise TypeError('given arg must be numeric')
            if isinstance(tolerance, numbers.Number) is False:
                raise TypeError('given tolerance arg must be numeric')
            if tolerance < 0:
                raise ValueError('given tolerance arg must be positive')

    def _check_dict_like(self, d, check_keys=True, check_values=True, check_getitem=True, name='val', return_as_bool=False):
        """Helper to check if given val has various dict-like attributes."""
        if not isinstance(d, Iterable):
            if return_as_bool:
                return False
            else:
                raise TypeError('%s <%s> is not dict-like: not iterable' % (name, type(d).__name__))
        if check_keys:
            if not hasattr(d, 'keys') or not callable(getattr(d, 'keys')):
                if return_as_bool:
                    return False
                else:
                    raise TypeError('%s <%s> is not dict-like: missing keys()' % (name, type(d).__name__))
        if check_values:
            if not hasattr(d, 'values') or not callable(getattr(d, 'values')):
                if return_as_bool:
                    return False
                else:
                    raise TypeError('%s <%s> is not dict-like: missing values()' % (name, type(d).__name__))
        if check_getitem:
            if not hasattr(d, '__getitem__'):
                if return_as_bool:
                    return False
                else:
                    raise TypeError('%s <%s> is not dict-like: missing [] accessor' % (name, type(d).__name__))
        if return_as_bool:
            return True

    def _check_iterable(self, l, check_getitem=True, name='val'):
        """Helper to check if given val has various iterable attributes."""
        if not isinstance(l, Iterable):
            raise TypeError('%s <%s> is not iterable' % (name, type(l).__name__))
        if check_getitem:
            if not hasattr(l, '__getitem__'):
                raise TypeError('%s <%s> does not have [] accessor' % (name, type(l).__name__))

    def _dict_not_equal(self, val, other, ignore=None, include=None):
        """Helper to compare dicts."""
        if ignore or include:
            ignores = self._dict_ignore(ignore)
            includes = self._dict_include(include)

            # guarantee include keys are in val
            if include:
                missing = []
                for i in includes:
                    if i not in val:
                        missing.append(i)
                if missing:
                    self.error('Expected <%s> to include key%s %s, but did not include key%s %s.' % (
                        val,
                        '' if len(includes) == 1 else 's',
                        self._fmt_items(['.'.join([str(s) for s in i]) if type(i) is tuple else i for i in includes]),
                        '' if len(missing) == 1 else 's',
                        self._fmt_items(missing)))

            # calc val keys given ignores and includes
            if ignore and include:
                k1 = set([k for k in val if k not in ignores and k in includes])
            elif ignore:
                k1 = set([k for k in val if k not in ignores])
            else:  # include
                k1 = set([k for k in val if k in includes])

            # calc other keys given ignores and includes
            if ignore and include:
                k2 = set([k for k in other if k not in ignores and k in includes])
            elif ignore:
                k2 = set([k for k in other if k not in ignores])
            else:  # include
                k2 = set([k for k in other if k in includes])

            if k1 != k2:
                # different set of keys, so not equal
                return True
            else:
                for k in k1:
                    if self._check_dict_like(val[k], check_values=False, return_as_bool=True) and \
                            self._check_dict_like(other[k], check_values=False, return_as_bool=True):
                        subdicts_not_equal = self._dict_not_equal(
                            val[k],
                            other[k],
                            ignore=[i[1:] for i in ignores if type(i) is tuple and i[0] == k] if ignore else None,
                            include=[i[1:] for i in self._dict_ignore(include) if type(i) is tuple and i[0] == k] if include else None)
                        if subdicts_not_equal:
                            # fast fail inside the loop since sub-dicts are not equal
                            return True
                    elif val[k] != other[k]:
                        # fast fail inside the loop since values are not equal
                        return True
            return False
        else:
            return val != other

    def _dict_ignore(self, ignore):
        """Helper to make list for given ignore kwarg values."""
        return [i[0] if type(i) is tuple and len(i) == 1 else i for i in (ignore if type(ignore) is list else [ignore])]

    def _dict_include(self, include):
        """Helper to make a list from given include kwarg values."""
        return [i[0] if type(i) is tuple else i for i in (include if type(include) is list else [include])]

    def _dict_err(self, val, other, ignore=None, include=None):
        """Helper to construct error message for dict comparison."""
        def _dict_repr(d, other):
            out = ''
            ellip = False
            for k, v in sorted(d.items()):
                if k not in other:
                    out += '%s%s: %s' % (', ' if len(out) > 0 else '', repr(k), repr(v))
                elif v != other[k]:
                    out += '%s%s: %s' % (
                        ', ' if len(out) > 0 else '',
                        repr(k),
                        _dict_repr(v, other[k]) if self._check_dict_like(
                            v, check_values=False, return_as_bool=True) and self._check_dict_like(
                                other[k], check_values=False, return_as_bool=True) else repr(v)
                    )
                else:
                    ellip = True
            return '{%s%s}' % ('..' if ellip and len(out) == 0 else '.., ' if ellip else '', out)

        if ignore:
            ignores = self._dict_ignore(ignore)
            ignore_err = ' ignoring keys %s' % self._fmt_items(['.'.join([str(s) for s in i]) if type(i) is tuple else i for i in ignores])
        if include:
            includes = self._dict_ignore(include)
            include_err = ' including keys %s' % self._fmt_items(['.'.join([str(s) for s in i]) if type(i) is tuple else i for i in includes])

        self.error('Expected <%s> to be equal to <%s>%s%s, but was not.' % (
            _dict_repr(val, other),
            _dict_repr(other, val),
            ignore_err if ignore else '',
            include_err if include else ''
        ))



import locale
import os
import re
import shutil
import subprocess
from pathlib import Path

from _pytest.fixtures import fixture
from assertpy import assert_that
from freezegun import freeze_time

from bashckup.bashckup import main

current_path = Path(os.path.dirname(os.path.realpath(__file__)))
tests_path = current_path / '..' / '..'
conf_path = tests_path / 'resources' / 'confs'
locale.setlocale(locale.LC_ALL, 'en_US.UTF-8')


@freeze_time('2023-07-10 15:02:10')
def test_mariadb(backup_folder):
    """
    GOAL: Test mariadb dump
    """
    # Given
    config_file = conf_path / 'mariadb.yml'
    expected_backup_folder = backup_folder / 'mariadb'
    # When
    return_code = main(['backup', 'file', '--config-file', str(config_file)])

    # Then
    assert_that(return_code).is_equal_to(0)
    output = []
    with os.scandir(expected_backup_folder) as it:
        entry: os.DirEntry
        for entry in it:
            output.append({'file-name': entry.name})

    assert_that(output).contains_only({'file-name': '2023-07-10T15:02:10-mariadb.sql'})

    # Check if dump is equal to what are inside the DB
    # Remove COLLATE because it's not present in same way one a newer version of mariadb
    actual = (expected_backup_folder / '2023-07-10T15:02:10-mariadb.sql').read_text().replace('COLLATE utf8mb4_bin ',
                                                                                              '')
    expected = (tests_path / 'resources' / 'mariadb' / 'sqlDump.sql').read_text().replace('COLLATE utf8mb4_bin ', '')
    assert_that(actual).contains(expected)



import locale
import os

from assertpy import assert_that
from pathlib import Path

from bashckup.actuators.readers import FileReader
from bashckup.actuators.writers import FileWriter

current_path = Path(os.path.dirname(os.path.realpath(__file__)))
tests_path = current_path / '..'
conf_path = tests_path / 'resources' / 'confs'
locale.setlocale(locale.LC_ALL, 'en_US.UTF-8')

"""
Change working directory to test folder
"""


def test_generate_dry_run_backup_cmd(backup_folder, server_data_folder):
    # Given
    global_context = {'backup-id': 'test', 'dry-run': True, 'verbose': True, 'backup': True}
    args = {'path': str(server_data_folder)}
    metadata = {}
    reader_module = FileReader(global_context, args, metadata)
    writer_module = FileWriter(global_context, {'file-name': 'test', 'path': str(backup_folder)}, metadata)

    # When
    metadata.update(reader_module.prepare_module())
    metadata.update(writer_module.prepare_module())
    result = reader_module.generate_dry_run_backup_cmd()

    # Then
    assert_that(result).is_length(4)
    assert_that(result[0]).is_equal_to('tar')
    assert_that(result[1]).is_equal_to('--verbose')
    assert_that(result[2]).is_equal_to('--create')
    assert_that(result[3]).ends_with('tests/serverData')



import os
import shutil
import subprocess
from abc import ABC
from datetime import datetime
from pathlib import Path
from typing import IO, AnyStr, Dict

from jsonschema.validators import validate

from bashckup.actuators import writers
from bashckup.actuators.actuators import ActuatorMetadata, CommandActuator
from bashckup.actuators.exceptions import ParameterException


class AbstractReader(CommandActuator, ABC):

    def _actuator_type(self) -> str:
        return 'reader'

    def generate_backup_process(self, stdin: IO[AnyStr] = None, stdout: IO[AnyStr] = subprocess.PIPE) \
            -> subprocess.Popen:
        return super().generate_backup_process(stdin, stdout)


class FileReader(AbstractReader):
    defaultLevel0frequency = 'weekly'
    validation_schema = {'type': 'object',
                         'properties': {
                             'path': {
                                 'type': 'string',
                                 'description': 'Folder path for the backup. It will create a sub-folder with the '
                                                'backup id'},
                             'incremental-metadata-file-prefix': {
                                 'type': 'string',
                                 'description': 'Name of metadata-file used to store difference between backups'},
                             'level-0-frequency': {
                                 'type': 'string',
                                 'enum': ['weekly', 'monthly'],
                                 'default': defaultLevel0frequency,
                                 'description': 'When a full backup have to be done ? You have to choose in ['
                                                '\'weekly\', \'monthly\']'}
                         },
                         'required': ['path'],
                         'additionalProperties': False}

    def __init__(self, global_context: dict, args: dict, metadata: Dict[str, Dict[str, ActuatorMetadata]] = None):
        super().__init__(global_context, args, metadata)
        self._file_prefix = None
        self._output_directory = None

    @staticmethod
    def module_name() -> str:
        return 'files'

    def _get_params(self) -> None:
        validate(self._args, self.validation_schema)
        if self._args.get('level-0-frequency') is not None and self._args.get(
                'incremental-metadata-file-prefix') is None:
            raise ParameterException('level-0-frequency can be used only with incremental-metadata-file-prefix',
                                     'level-0-frequency', self._backup_id, self.module_name())

        self.incrementalMetadataFilePrefix = self._args.get('incremental-metadata-file-prefix')
        self.path = self._args['path']
        self.level0Frequency = self._args.get('level-0-frequency', self.defaultLevel0frequency)

    def _generate_backup_cmd(self) -> [str]:
        # Validate metadata
        self._validate_register_metadata()

        cmd = ['tar']
        if self._verbose:
            cmd.append('--verbose')
        if self.incrementalMetadataFilePrefix is not None:
            cmd.extend(['--listed-incremental', str(self._generate_incremental_metadata_file_name())])
        cmd.extend(['--create', self.path])
        return cmd

    def _validate_register_metadata(self):
        """
        Validates metadata and register attributes with it
        """
        output_directories = [v.get('output-directory') for (i, v) in self._metadata['writer'].items()]
        if len(output_directories) != 1:  # Because we need at least one, and it can not be greater than 1
            raise ValueError('output-directory must be defined in writer module')
        self._output_directory = output_directories[0]
        file_prefixes = [v.get('file-prefix') for (i, v) in self._metadata['writer'].items()]
        if len(file_prefixes) != 1:  # Because we need at least one, and it can not be greater than 1
            raise ValueError('file-prefix must be defined in writer module')
        self._file_prefix = file_prefixes[0]
        backups_datetime = [v.get('backup-datetime') for (i, v) in self._metadata['writer'].items()]
        if len(backups_datetime) != 1:  # Because we need at least one, and it can not be greater than 1
            raise ValueError('backup-datetime must be defined in writer module')
        self._backup_datetime = datetime.fromisoformat(backups_datetime[0])

    def _generate_incremental_metadata_file_name(self) -> Path:
        # self._file_prefix
        file_name = self.incrementalMetadataFilePrefix
        if self.level0Frequency == 'weekly':
            file_name += '-w' + self._backup_datetime.strftime('%V')
        elif self.level0Frequency == 'monthly':
            file_name += '-m' + self._backup_datetime.strftime('%m')
        else:
            raise ValueError(f'Frequency {self.level0Frequency} is not managed')
        file_name += '.snar'

        # Search if an incremental file already exists
        incremental_file_exists = False
        # Folder can be absent only in dry mode
        if not self._dry_run and Path(self._output_directory).is_dir():
            with os.scandir(self._output_directory) as it:
                entry: os.DirEntry
                for entry in it:
                    if entry.name.endswith(file_name):
                        file_name = entry.name
                        incremental_file_exists = True
                        break
        # If incremental file does not exist then create a file with the prefix
        if not incremental_file_exists:
            file_name = self._file_prefix + file_name
        return Path(self._output_directory) / file_name

    def _generate_metadata(self) -> dict:
        days = 0
        # Incremental backup, so we need to keep level0 backup
        if self._isBackup and self.incrementalMetadataFilePrefix is not None:
            if self.level0Frequency == 'weekly':
                # We can use datetime.today() instead of datetime set by writer because we add 1 day of margin
                # And in this part of code we can not have datetime set by writer
                week_number = datetime.today().strftime('%Y/%W')
                window_first_day = datetime.strptime(week_number + '/1', '%Y/%W/%w')  # Year/week-number/day-of-week
                days = (datetime.today() - window_first_day).days + 1
                # +1 because we want to protect all the day, not actual time of the first day of the week
            elif self.level0Frequency == 'monthly':
                month_number = datetime.today().strftime('%Y/%m/01')
                window_first_day = datetime.strptime(month_number, '%Y/%m/%d')  # Year/month/day-of-month
                days = (datetime.today() - window_first_day).days + 1
                # +1 because we want to protect all the day, not actual time of the first day of the month
            else:
                raise ValueError(f'Frequency {self.level0Frequency} is not managed')
        return {'file-preservation-window': days}

    def _generate_restore_cmd(self) -> [str]:
        # Validate metadata
        self._validate_register_metadata()
        cmd = ['tar']
        if self._verbose:
            cmd.append('--verbose')
        if self.incrementalMetadataFilePrefix is not None:
            cmd.extend(['--listed-incremental', str(self._generate_incremental_metadata_file_name())])
        cmd.extend(
            ['--extract', '--same-owner', '--same-permissions', '--strip-components=1', '--directory', self.path])

        return cmd

    # Override because we need to back up files before the restoration
    def generate_restore_process(self, stdin: IO[AnyStr] = subprocess.PIPE,
                                 stdout: IO[AnyStr] = None) -> subprocess.Popen:
        # Back up src files before restoration
        src_path = Path(self.path)
        files = os.listdir(src_path)
        if len(files) != 0:  # Not Empty
            backup_path = src_path.parents[0] / (
                    src_path.name + '-bck-' + datetime.today().isoformat(timespec='seconds'))

            metadata = os.stat(src_path)
            os.mkdir(backup_path, metadata.st_mode)
            for file in files:
                shutil.move(os.path.join(src_path, file), backup_path)
        return super().generate_restore_process(stdin, stdout)


class MariaDBReader(AbstractReader):
    validation_schema = {'type': 'object',
                         'properties': {
                             'database-name': {
                                 'type': 'string',
                                 'description': 'Name of mariadb database'}
                         },
                         'required': ['database-name'],
                         'additionalProperties': False}

    @staticmethod
    def module_name() -> str:
        return 'mariaDBDatabase'

    def _get_params(self) -> None:
        validate(self._args, self.validation_schema)

        self.databaseName = self._args['database-name']

    def _generate_backup_cmd(self) -> [str]:
        cmd = ['mysqldump']
        if self._verbose:
            cmd.append('--verbose')
        cmd.append(self.databaseName)
        return cmd

    def _generate_restore_cmd(self) -> [str]:
        cmd = ['mysql']
        if self._verbose:
            cmd.append('--verbose')
        cmd.append(self.databaseName)
        return cmd



import locale
import os
import shutil
from pathlib import Path

from assertpy import assert_that
from freezegun import freeze_time

from bashckup.bashckup import main

current_path = Path(os.path.dirname(os.path.realpath(__file__)))
tests_path = current_path / '..' / '..'
conf_path = tests_path / 'resources' / 'confs'
files_path = tests_path / 'resources' / 'tar'
locale.setlocale(locale.LC_ALL, 'en_US.UTF-8')


@freeze_time('2023-07-10 15:02:10')
def test_tar(backup_folder, server_data_folder):
    """
    Goal: Test tar without incremental options
    """
    # Given
    config_file = conf_path / 'tar.yml'
    expected_backup_folder = backup_folder / 'tar'
    # When
    return_code = main(['backup', 'file', '--config-file', str(config_file)])

    # Then
    assert_that(return_code).is_equal_to(0)
    output = []
    with os.scandir(expected_backup_folder) as it:
        entry: os.DirEntry
        for entry in it:
            output.append({'file-name': entry.name, 'size': entry.stat().st_size})

    assert_that(output).contains_only({'file-name': '2023-07-10T15:02:10-tar.tar', 'size': 10240})


@freeze_time('2023-07-10 15:02:10')
def test_incremental_tar(backup_folder, server_data_folder):
    """
    GOAL: Test TAR with incremental options, first backup
    """
    # Given
    config_file = conf_path / 'tar-diff.yml'
    expected_backup_folder = backup_folder / 'tar-diff'
    # When
    return_code = main(['backup', 'file', '--config-file', str(config_file)])

    # Then
    assert_that(return_code).is_equal_to(0)
    output = []
    with os.scandir(expected_backup_folder) as it:
        entry: os.DirEntry
        for entry in it:
            output.append({'file-name': entry.name, 'size': entry.stat().st_size})

    assert_that(output).is_length(2)
    snar_file = list(filter(lambda f: f['file-name'].endswith('.snar'), output))[0]
    bck_file = list(filter(lambda f: f['file-name'].endswith('.tar'), output))[0]
    assert_that(snar_file['file-name']).is_equal_to('2023-07-10T15:02:10-tar-snap-w28.snar')
    assert_that(snar_file['size']).is_between(80, 100)

    assert_that(bck_file['file-name']).is_equal_to('2023-07-10T15:02:10-tar-diff.tar')
    assert_that(bck_file['size']).is_equal_to(10240)


@freeze_time('2023-07-10 15:02:10')
def test_incremental_tar_incremental_backup(backup_folder, server_data_folder):
    """
    GOAL: Test TAR with incremental options, incremental backup. File 2 is not present in initial archive
    """
    # Given
    config_file = conf_path / 'tar-diff.yml'
    expected_backup_folder = backup_folder / 'tar-diff'
    os.mkdir(expected_backup_folder)
    shutil.copyfile(files_path / '2023-07-09T12:03:34-tar-snap-w28.snar',
                    expected_backup_folder / '2023-07-09T12:03:34-tar-snap-w28.snar')
    shutil.copyfile(files_path / '2023-07-09T12:03:34-tar-diff.tar',
                    expected_backup_folder / '2023-07-09T12:03:34-tar-diff.tar')
    # When
    return_code = main(['backup', 'file', '--config-file', str(config_file)])

    # Then
    assert_that(return_code).is_equal_to(0)
    output = []
    with os.scandir(expected_backup_folder) as it:
        entry: os.DirEntry
        for entry in it:
            output.append({'file-name': entry.name})

    assert_that(output).is_length(3)
    assert_that(output).contains_only(
        {'file-name': '2023-07-09T12:03:34-tar-snap-w28.snar'},
        {'file-name': '2023-07-09T12:03:34-tar-diff.tar'},
        {'file-name': '2023-07-10T15:02:10-tar-diff.tar'}
    )



import logging
import os
import re
import subprocess
from abc import ABC
from datetime import datetime
from pathlib import Path
from typing import IO, AnyStr, Any

from jsonschema.validators import validate

from bashckup.actuators.actuators import CommandActuator
from bashckup.actuators.exceptions import ParameterException, ModuleException

datetimeOutputFileRegex = re.compile(r'^(\d+-\d+-\d+T\d+:\d+:\d+)')
outputFileRegex = re.compile(r'^(\d+-\d+-\d+T\d+:\d+:\d+)-(.*)')


class AbstractWriter(CommandActuator, ABC):

    def _actuator_type(self) -> str:
        return 'writer'

    def generate_backup_process(self, stdin: IO[AnyStr], stdout: IO[AnyStr] = subprocess.DEVNULL) \
            -> subprocess.Popen:
        return super().generate_backup_process(stdin, stdout)

    def generate_restore_process(self, stdin: IO[AnyStr] = None, stdout: IO[AnyStr] = subprocess.PIPE) \
            -> subprocess.Popen:
        return super().generate_restore_process(stdin, stdout)


class FileWriter(AbstractWriter):
    validation_schema = {'type': 'object',
                         'properties': {
                             'path': {
                                 'type': 'string',
                                 'description': 'Path to the output folder'},
                             'file-name': {
                                 'type': 'string',
                                 'description': 'File name of the backup file'}
                         },
                         'required': ['path', 'file-name'],
                         'additionalProperties': False}

    @staticmethod
    def module_name() -> str:
        return 'outputFile'

    def _get_params(self) -> None:
        validate(self._args, self.validation_schema)

        self.path = Path(self._args['path'])
        self.file_name = self._args['file-name']
        self._output_folder = self.path

        # If folder doesn't exist, it will be created by _pre_run_tasks
        if self._output_folder.exists() and not self._output_folder.is_dir():
            raise ParameterException(f'''Directory [{self._output_folder}] is NOT a directory''', 'path',
                                     self._backup_id, self.module_name())
        if self._output_folder.exists() and not os.access(self._output_folder, os.W_OK):
            raise ParameterException(f'''Directory [{self._output_folder}] is not writable''', 'path',
                                     self._backup_id, self.module_name())

    def _generate_backup_cmd(self) -> [str]:
        cmd = ['cat']
        return cmd

    # Override because this module have a special way to managed process
    def generate_dry_run_backup_cmd(self) -> [str]:
        return ['>', str(self._output_file_path)]

    # Override because this module have a special way to managed process
    def generate_backup_process(self, stdin: IO[AnyStr], stdout: IO[AnyStr] = None) -> subprocess.Popen:
        try:
            file_descriptor = os.open(self._output_file_path, flags=os.O_WRONLY | os.O_CREAT | os.O_EXCL, mode=0o600)
            with open(file_descriptor, 'w') as f:
                return super().generate_backup_process(stdin, f)
        except FileExistsError:
            raise ModuleException(f'File [{self._output_file_path}] already exists', self._backup_id,
                                  self.module_name())

    def _pre_run_tasks(self) -> None:
        #  Create backup folder
        if not os.path.exists(self._output_folder) and self._dry_run is False:
            logging.info('Folder [%s] created', str(self._output_folder))
            os.makedirs(self._output_folder, mode=0o750)
        elif not os.path.exists(self._output_folder) and self._dry_run is True:
            logging.info('Folder [%s] would have been created.', self._output_folder)
        if self._isBackup:
            self._backup_datetime = datetime.today()
            self._file_prefix = self._backup_datetime.isoformat(timespec='seconds') + '-'
            self._output_file_path = self._output_folder / (self._file_prefix + self.file_name)
        else:
            latest_backup = self._get_latest_backup_file()
            self._backup_datetime = latest_backup['backup-datetime']
            self._file_prefix = self._backup_datetime.isoformat(timespec='seconds') + '-'
            self._output_file_path = latest_backup['file-path']

    def _generate_metadata(self) -> dict:
        return {'output-directory': str(self._output_folder), 'file-prefix': self._file_prefix,
                'backup-datetime': self._backup_datetime.isoformat(timespec='seconds')}

    def _get_latest_backup_file(self) -> {Any}:
        latest_backup_path = None
        latest_backup_datetime = None
        with os.scandir(self._output_folder) as it:
            entry: os.DirEntry
            for entry in it:
                matches = outputFileRegex.search(entry.name)
                if matches is None or len(matches.groups()) != 2:
                    continue
                if matches.group(2) != self.file_name:
                    continue
                current_backup_datetime = datetime.fromisoformat(matches.group(1))
                if latest_backup_datetime is None or latest_backup_datetime < current_backup_datetime:
                    latest_backup_datetime = current_backup_datetime
                    latest_backup_path = entry.path
        if latest_backup_path is None:
            raise ParameterException(f'''path [{str(self._output_folder)}] does not contains any backup''',
                                     'path', self._backup_id, self.module_name())

        return {'file-path': latest_backup_path, 'backup-datetime': latest_backup_datetime}

    def _generate_restore_cmd(self) -> [str]:
        cmd = ['cat', self._output_file_path]

        return cmd



import locale
import logging
import os
import shutil
from pathlib import Path

import pytest
from _pytest.fixtures import fixture
from assertpy import assert_that
from freezegun import freeze_time

from bashckup.bashckup import main

current_path = Path(os.path.dirname(os.path.realpath(__file__)))
tests_path = current_path / '..' / '..'
conf_path = tests_path / 'resources' / 'confs'
locale.setlocale(locale.LC_ALL, 'en_US.UTF-8')

"""
Depends on TAR tests
"""


@fixture
def change_crypt_password_file_rights():
    """ Change password file right and the reset to default """
    crypt_password_file = conf_path / 'crypt-password.pwd'
    file_stat = os.stat(crypt_password_file)
    file_mod = file_stat.st_mode
    file_owner = file_stat.st_uid
    file_group = file_stat.st_gid
    os.chmod(crypt_password_file, 0o600)
    os.chown(crypt_password_file, 0, 0)
    yield crypt_password_file
    os.chmod(crypt_password_file, file_mod)
    os.chown(crypt_password_file, file_owner, file_group)


@fixture
def change_crypt_password_file_with_good_rights_and_bad_owner():
    """ Change password file right and the reset to default """
    crypt_password_file = conf_path / 'crypt-password.pwd'
    file_stat = os.stat(crypt_password_file)
    file_mod = file_stat.st_mode
    file_owner = file_stat.st_uid
    file_group = file_stat.st_gid
    os.chmod(crypt_password_file, 0o600)
    os.chown(crypt_password_file, os.getuid() + 1, 0)
    yield crypt_password_file
    os.chmod(crypt_password_file, file_mod)
    os.chown(crypt_password_file, file_owner, file_group)


@fixture
def change_crypt_password_file_readable_for_group_and_good_owner():
    """ Change password file right and the reset to default """
    crypt_password_file = conf_path / 'crypt-password.pwd'
    file_stat = os.stat(crypt_password_file)
    file_mod = file_stat.st_mode
    file_owner = file_stat.st_uid
    file_group = file_stat.st_gid
    os.chmod(crypt_password_file, 0o660)
    os.chown(crypt_password_file, os.getuid(), 0)
    yield crypt_password_file
    os.chmod(crypt_password_file, file_mod)
    os.chown(crypt_password_file, file_owner, file_group)


@fixture
def change_crypt_password_file_readable_for_everyone_and_good_owner():
    """ Change password file right and the reset to default """
    crypt_password_file = conf_path / 'crypt-password.pwd'
    file_stat = os.stat(crypt_password_file)
    file_mod = file_stat.st_mode
    file_owner = file_stat.st_uid
    file_group = file_stat.st_gid
    os.chmod(crypt_password_file, 0o666)
    os.chown(crypt_password_file, os.getuid(), 0)
    yield crypt_password_file
    os.chmod(crypt_password_file, file_mod)
    os.chown(crypt_password_file, file_owner, file_group)


@freeze_time('2023-07-10 15:02:10')
def test_tar_crypt_password_readable_for_everyone(caplog,
                                                  change_crypt_password_file_readable_for_everyone_and_good_owner,
                                                  backup_folder, server_data_folder):
    """
    Goal: Test file permission check on password file
    """
    caplog.set_level(logging.DEBUG)
    # Given
    config_file = conf_path / 'tar-crypt.yml'
    # When
    with pytest.raises(SystemExit) as e:
        main(['backup', 'file', '--config-file', str(config_file)])

    # Then
    assert e.type == SystemExit
    assert e.value.code == 1
    assert_that(caplog.record_tuples).contains(
        ('root', logging.ERROR, 'ERROR: On the [tar-crypt] backup\n'
                                'On the [crypt] module\n'
                                'Validation error on parameter: password-file\n'
                                'Reason: File [resources/confs/crypt-password.pwd] must not be readable from group '
                                'and from others'))


@freeze_time('2023-07-10 15:02:10')
def test_tar_crypt_password_owned_by_another_user(caplog, change_crypt_password_file_with_good_rights_and_bad_owner,
                                                  backup_folder, server_data_folder):
    """
    Goal: Test file owner check on password file
    """
    caplog.set_level(logging.DEBUG)
    # Given
    config_file = conf_path / 'tar-crypt.yml'
    # When
    with pytest.raises(SystemExit) as e:
        main(['backup', 'file', '--config-file', str(config_file)])

    # Then
    assert e.type == SystemExit
    assert e.value.code == 1
    assert_that(caplog.record_tuples).contains(
        ('root', logging.ERROR, 'ERROR: On the [tar-crypt] backup\n'
                                'On the [crypt] module\n'
                                'Validation error on parameter: password-file\n'
                                'Reason: File [resources/confs/crypt-password.pwd] is not owned by current '
                                'user'))


@freeze_time('2023-07-10 15:02:10')
def test_tar_crypt_password_readable_for_group(caplog, change_crypt_password_file_readable_for_group_and_good_owner,
                                               backup_folder, server_data_folder):
    """
    Goal: Test file permission check on password file for group
    """
    caplog.set_level(logging.DEBUG)
    # Given
    config_file = conf_path / 'tar-crypt.yml'
    # When
    with pytest.raises(SystemExit) as e:
        main(['backup', 'file', '--config-file', str(config_file)])

    # Then
    assert e.type == SystemExit
    assert e.value.code == 1
    assert_that(caplog.record_tuples).contains(
        ('root', logging.ERROR, 'ERROR: On the [tar-crypt] backup\n'
                                'On the [crypt] module\n'
                                'Validation error on parameter: password-file\n'
                                'Reason: File [resources/confs/crypt-password.pwd] must not be readable from group '
                                'and from others'))


@freeze_time('2023-07-10 15:02:10')
def test_tar_crypt_password(backup_folder, server_data_folder, change_crypt_password_file_rights):
    """
    GOAL: Test CRYPT
    """
    # Given
    config_file = conf_path / 'tar-crypt.yml'
    expected_backup_folder = backup_folder / 'tar-crypt'
    # When
    return_code = main(['backup', 'file', '--config-file', str(config_file)])

    # Then
    assert_that(return_code).is_equal_to(0)
    output = []
    with os.scandir(expected_backup_folder) as it:
        entry: os.DirEntry
        for entry in it:
            output.append({'file-name': entry.name, 'size': entry.stat().st_size})

    assert_that(output).is_length(1)
    bck_file = output[0]
    assert_that(bck_file['file-name']).is_equal_to('2023-07-10T15:02:10-tar-crypt.tar.crypt')
    # Info: Without encryption its 10240 octets
    assert_that(bck_file['size']).is_between(10240, 10280)

    actual = (expected_backup_folder / '2023-07-10T15:02:10-tar-crypt.tar.crypt').read_text(errors='ignore')
    assert_that(actual).starts_with('Salted__')



import locale
import os
import shutil
from pathlib import Path

import pytest
from assertpy import assert_that
from freezegun import freeze_time

from bashckup.bashckup import main

current_path = Path(os.path.dirname(os.path.realpath(__file__)))
tests_path = current_path / '..' / '..'
conf_path = tests_path / 'resources' / 'confs'
files_path = tests_path / 'resources' / 'tar'
server_data_folder = tests_path / 'serverData'
locale.setlocale(locale.LC_ALL, 'en_US.UTF-8')


@freeze_time('2023-07-10 15:02:10')
def test_restore_tar(backup_folder):
    """
    GOAL: 2023-07-10T15:02:10-tar.tar fil be used
    2023-07-10T15:02:12-unknownBackup.tar is more recent but not with the correct backup file name
    """
    # Given
    config_file = conf_path / 'tar.yml'
    expected_backup_folder = backup_folder / 'tar'
    try:
        os.makedirs(expected_backup_folder)
        os.makedirs(server_data_folder)

        shutil.copyfile(files_path / '2023-07-10T15:02:10-tar.tar',
                        expected_backup_folder / '2023-07-10T15:02:10-tar.tar')
        shutil.copyfile(files_path / '2023-07-10T15:02:12-unknownBackup.tar',
                        expected_backup_folder / '2023-07-10T15:02:12-unknownBackup.tar')
        # When
        return_code = main(['restore', 'file', '--config-file', str(config_file)])
        # return_code = main(['--dry-run', 'restore', 'file', '--config-file', str(config_file)])

        # Then
        assert_that(return_code).is_equal_to(0)
        output = []
        with os.scandir(server_data_folder) as it:
            entry: os.DirEntry
            for entry in it:
                output.append({'file-name': entry.name, 'size': entry.stat().st_size})

        assert_that(output).contains_only({'file-name': 'file1', 'size': 17}, {'file-name': 'file2', 'size': 17})
    finally:
        shutil.rmtree(server_data_folder, ignore_errors=True)


@freeze_time('2023-07-10 15:02:10')
def test_restore_tar_backup_folder_not_empty(backup_folder):
    """
    GOAL: Previous files are copied in a folder and restoration is done
    """
    # Given

    config_file = conf_path / 'tar.yml'
    expected_backup_folder = backup_folder / 'tar'
    backup_of_backup_folder = str(server_data_folder) + '-bck-2023-07-10T15:02:10'

    try:
        os.mkdir(expected_backup_folder)
        shutil.copyfile(files_path / '2023-07-10T15:02:10-tar.tar',
                        expected_backup_folder / '2023-07-10T15:02:10-tar.tar')
        os.makedirs(server_data_folder)
        with open(server_data_folder / 'brokenBackup.txt', "w") as file:
            file.write("This file has to be backup when restoration will be done")
        # When
        return_code = main(['restore', 'file', '--config-file', str(config_file)])

        # Then
        assert_that(return_code).is_equal_to(0)

        # Check backupFolder
        output = []
        with os.scandir(server_data_folder) as it:
            entry: os.DirEntry
            for entry in it:
                output.append({'file-name': entry.name, 'size': entry.stat().st_size})

        assert_that(output).contains_only({'file-name': 'file1', 'size': 17}, {'file-name': 'file2', 'size': 17})

        # Check backupFolder-bck
        output = []
        with os.scandir(backup_of_backup_folder) as it:
            entry: os.DirEntry
            for entry in it:
                output.append({'file-name': entry.name, 'size': entry.stat().st_size})

        assert_that(output).contains_only({'file-name': 'brokenBackup.txt', 'size': 56})
    finally:
        shutil.rmtree(backup_of_backup_folder, ignore_errors=True)
        shutil.rmtree(server_data_folder, ignore_errors=True)


@pytest.mark.skip(reason="Restoration requires a deep modification of the architecture, it will be done for the next "
                         "version")
@freeze_time('2023-07-10 15:02:10')
def test_restore_incremental_tar_incremental_backup(backup_folder):
    """
    GOAL: Test restore a incremental backup with an increment
    """
    # Given
    config_file = conf_path / 'tar-diff.yml'
    expected_backup_folder = backup_folder / 'tar-diff'
    try:
        shutil.copytree(files_path / 'incremental', expected_backup_folder)
        os.makedirs(server_data_folder)

        # When
        return_code = main(['backup', 'file', '--config-file', str(config_file)])

        # Then
        assert_that(return_code).is_equal_to(0)
        output = []
        with os.scandir(server_data_folder) as it:
            entry: os.DirEntry
            for entry in it:
                output.append({'file-name': entry.name})

        assert_that(output).contains_only({'file-name': 'file1', 'size': 17}, {'file-name': 'file2', 'size': 17})
    finally:
        shutil.rmtree(server_data_folder)



import subprocess
from abc import abstractmethod
from typing import AnyStr, IO, Dict

from jsonschema.exceptions import ValidationError
from jsonschema.validators import validate

from bashckup.actuators.exceptions import ParameterException


class ActuatorMetadata:
    _schema = {'type': 'object',
               'properties': {
                   'file-preservation-window': {
                       'type': 'integer',
                       'minimum': 0,
                       'description': 'How many days files have to stay, it override cleaner\'s specification.'},
                   'output-directory': {
                       'type': 'string',
                       'description': 'Path to backup folder'},
                   'file-prefix': {
                       'type': 'string',
                       'description': 'Prefix of each files (with timestamp)'},
                   'backup-datetime': {
                       'type': 'string',
                       'format': 'date-time',
                       'description': 'Date time of the backup'}
               },
               'additionalProperties': False
               }

    def __init__(self, metadata: dict):
        validate(metadata, self._schema)
        self.__dict__.update(metadata)

    def get(self, key):
        return self.__dict__.get(key)


class AbstractActuator:
    def __init__(self, global_context: dict, args: dict, metadata: Dict[str, Dict[str, ActuatorMetadata]] = None):
        self._backup_id = global_context['backup-id']
        self._dry_run = global_context['dry-run']
        self._verbose = global_context['verbose']
        self._isBackup = global_context['backup']
        self._isRestore = not global_context['backup']
        self._args: dict = args
        self._metadata: Dict[str, Dict[str, ActuatorMetadata]] = metadata

    @staticmethod
    @abstractmethod
    def module_name() -> str:
        pass

    @abstractmethod
    def _get_params(self) -> None:
        pass

    @abstractmethod
    def _actuator_type(self) -> str:
        pass

    def _pre_run_tasks(self) -> None:
        """ If an actuator needs to create something before running the backup plan, you can use this method """
        pass

    def _generate_metadata(self) -> dict:
        return {}

    def _validate_parameters(self) -> None:
        try:
            self._get_params()
        except ValidationError as e:
            raise ParameterException(f'{e.message}\n'
                                     f'''Helper: {e.schema.get('description')}''', e.json_path, self._backup_id,
                                     self.module_name()) from None

    def prepare_module(self):
        """ Do validation then run pre task and finally init metadata """
        self._validate_parameters()
        self._pre_run_tasks()
        return {self._actuator_type(): {self.module_name(): ActuatorMetadata(self._generate_metadata())}}


class CommandActuator(AbstractActuator):
    @abstractmethod
    def _generate_backup_cmd(self) -> [str]:
        pass

    def generate_dry_run_backup_cmd(self) -> [str]:
        if self._dry_run is False:
            raise Exception('You are not allowed to call this function outside dry-run')
        return self._generate_backup_cmd()

    def generate_backup_process(self, stdin: IO[AnyStr], stdout: IO[AnyStr]) -> subprocess.Popen:
        if self._dry_run is True:
            raise Exception('You are not allowed to call this function in dry-run')
        return subprocess.Popen(self._generate_backup_cmd(), shell=False, stdin=stdin, stdout=stdout,
                                stderr=subprocess.PIPE)

    @abstractmethod
    def _generate_restore_cmd(self) -> [str]:
        pass

    def generate_dry_run_restore_cmd(self) -> [str]:
        if self._dry_run is False:
            raise Exception('You are not allowed to call this function outside dry-run')
        return self._generate_restore_cmd()

    def generate_restore_process(self, stdin: IO[AnyStr],
                                 stdout: IO[AnyStr] = None) -> subprocess.Popen:
        if self._dry_run is True:
            raise Exception('You are not allowed to call this function in dry-run')
        return subprocess.Popen(self._generate_restore_cmd(), shell=False, stdin=stdin, stdout=stdout,
                                stderr=subprocess.PIPE)


class PythonActuator(AbstractActuator):
    @abstractmethod
    def _prepare_run_backup(self) -> dict:
        """
        Compute all data useful for running backup
        This data can be used to run the backup or write the verbose
        :return: Dict of computed elements
        """
        pass

    @abstractmethod
    def _run_backup(self, args: dict) -> None:
        pass

    @abstractmethod
    def _dry_run_backup(self, args: dict) -> None:
        pass

    def run_backup(self) -> None:
        args = self._prepare_run_backup()
        if self._dry_run is True:
            self._dry_run_backup(args)
        else:
            self._run_backup(args)

    @abstractmethod
    def _prepare_run_restore(self) -> dict:
        """
        Compute all data useful for running restoration
        This data can be used to run the restoration or write the verbose
        :return: Dict of computed elements
        """
        pass

    @abstractmethod
    def _run_restore(self, args: dict) -> None:
        pass

    @abstractmethod
    def _dry_run_restore(self, args: dict) -> None:
        pass

    def run_restore(self) -> None:
        args = self._prepare_run_restore()
        if self._dry_run is True:
            self._dry_run_restore(args)
        else:
            self._run_restore(args)



# Copyright (c) 2015-2019, Activision Publishing, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import datetime

__tracebackhide__ = True


class DateMixin(object):
    """Date and time assertions mixin."""

    def is_before(self, other):
        """Asserts that val is a date and is before other date.

        Args:
            other: the other date, expected to be after val

        Examples:
            Usage::

                import datetime

                today = datetime.datetime.now()
                yesterday = today - datetime.timedelta(days=1)

                assert_that(yesterday).is_before(today)

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val is **not** before the given date
        """
        if type(self.val) is not datetime.datetime:
            raise TypeError('val must be datetime, but was type <%s>' % type(self.val).__name__)
        if type(other) is not datetime.datetime:
            raise TypeError('given arg must be datetime, but was type <%s>' % type(other).__name__)
        if self.val >= other:
            self.error('Expected <%s> to be before <%s>, but was not.' % (self.val.strftime('%Y-%m-%d %H:%M:%S'), other.strftime('%Y-%m-%d %H:%M:%S')))
        return self

    def is_after(self, other):
        """Asserts that val is a date and is after other date.

        Args:
            other: the other date, expected to be before val

        Examples:
            Usage::

                import datetime

                today = datetime.datetime.now()
                yesterday = today - datetime.timedelta(days=1)

                assert_that(today).is_after(yesterday)

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val is **not** after the given date
        """
        if type(self.val) is not datetime.datetime:
            raise TypeError('val must be datetime, but was type <%s>' % type(self.val).__name__)
        if type(other) is not datetime.datetime:
            raise TypeError('given arg must be datetime, but was type <%s>' % type(other).__name__)
        if self.val <= other:
            self.error('Expected <%s> to be after <%s>, but was not.' % (self.val.strftime('%Y-%m-%d %H:%M:%S'), other.strftime('%Y-%m-%d %H:%M:%S')))
        return self

    def is_equal_to_ignoring_milliseconds(self, other):
        """Asserts that val is a date and is equal to other date to the second.

        Args:
            other: the other date, expected to be equal to the second

        Examples:
            Usage::

                import datetime

                d1 = datetime.datetime(2020, 1, 2, 3, 4, 5, 6)       # 2020-01-02 03:04:05.000006
                d2 = datetime.datetime(2020, 1, 2, 3, 4, 5, 777777)  # 2020-01-02 03:04:05.777777

                assert_that(d1).is_equal_to_ignoring_milliseconds(d2)

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val is **not** equal to the given date to the second
        """
        if type(self.val) is not datetime.datetime:
            raise TypeError('val must be datetime, but was type <%s>' % type(self.val).__name__)
        if type(other) is not datetime.datetime:
            raise TypeError('given arg must be datetime, but was type <%s>' % type(other).__name__)
        if self.val.date() != other.date() or self.val.hour != other.hour or self.val.minute != other.minute or self.val.second != other.second:
            self.error('Expected <%s> to be equal to <%s>, but was not.' % (self.val.strftime('%Y-%m-%d %H:%M:%S'), other.strftime('%Y-%m-%d %H:%M:%S')))
        return self

    def is_equal_to_ignoring_seconds(self, other):
        """Asserts that val is a date and is equal to other date to the minute.

        Args:
            other: the other date, expected to be equal to the minute

        Examples:
            Usage::

                import datetime

                d1 = datetime.datetime(2020, 1, 2, 3, 4, 5)   # 2020-01-02 03:04:05
                d2 = datetime.datetime(2020, 1, 2, 3, 4, 55)  # 2020-01-02 03:04:55

                assert_that(d1).is_equal_to_ignoring_seconds(d2)

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val is **not** equal to the given date to the minute
        """
        if type(self.val) is not datetime.datetime:
            raise TypeError('val must be datetime, but was type <%s>' % type(self.val).__name__)
        if type(other) is not datetime.datetime:
            raise TypeError('given arg must be datetime, but was type <%s>' % type(other).__name__)
        if self.val.date() != other.date() or self.val.hour != other.hour or self.val.minute != other.minute:
            self.error('Expected <%s> to be equal to <%s>, but was not.' % (self.val.strftime('%Y-%m-%d %H:%M'), other.strftime('%Y-%m-%d %H:%M')))
        return self

    def is_equal_to_ignoring_time(self, other):
        """Asserts that val is a date and is equal to other date ignoring time.

        Args:
            other: the other date, expected to be equal ignoring time

        Examples:
            Usage::

                import datetime

                d1 = datetime.datetime(2020, 1, 2, 3, 4, 5)     # 2020-01-02 03:04:05
                d2 = datetime.datetime(2020, 1, 2, 13, 44, 55)  # 2020-01-02 13:44:55

                assert_that(d1).is_equal_to_ignoring_time(d2)

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val is **not** equal to the given date ignoring time
        """
        if type(self.val) is not datetime.datetime:
            raise TypeError('val must be datetime, but was type <%s>' % type(self.val).__name__)
        if type(other) is not datetime.datetime:
            raise TypeError('given arg must be datetime, but was type <%s>' % type(other).__name__)
        if self.val.date() != other.date():
            self.error('Expected <%s> to be equal to <%s>, but was not.' % (self.val.strftime('%Y-%m-%d'), other.strftime('%Y-%m-%d')))
        return self



from . import config
from ._async import wrap_coroutine
import copyreg
import dateutil
import datetime
import functools
import sys
import time
import uuid
import calendar
import unittest
import platform
import warnings
import types
import numbers
import inspect

from dateutil import parser
from dateutil.tz import tzlocal

try:
    from maya import MayaDT
except ImportError:
    MayaDT = None

_TIME_NS_PRESENT = hasattr(time, 'time_ns')
_MONOTONIC_NS_PRESENT = hasattr(time, 'monotonic_ns')
_PERF_COUNTER_NS_PRESENT = hasattr(time, 'perf_counter_ns')
_EPOCH = datetime.datetime(1970, 1, 1)
_EPOCHTZ = datetime.datetime(1970, 1, 1, tzinfo=dateutil.tz.UTC)

real_time = time.time
real_localtime = time.localtime
real_gmtime = time.gmtime
real_monotonic = time.monotonic
real_perf_counter = time.perf_counter
real_strftime = time.strftime
real_date = datetime.date
real_datetime = datetime.datetime
real_date_objects = [real_time, real_localtime, real_gmtime, real_monotonic, real_perf_counter, real_strftime, real_date, real_datetime]

if _TIME_NS_PRESENT:
    real_time_ns = time.time_ns
    real_date_objects.append(real_time_ns)

if _MONOTONIC_NS_PRESENT:
    real_monotonic_ns = time.monotonic_ns
    real_date_objects.append(real_monotonic_ns)

if _PERF_COUNTER_NS_PRESENT:
    real_perf_counter_ns = time.perf_counter_ns
    real_date_objects.append(real_perf_counter_ns)

_real_time_object_ids = {id(obj) for obj in real_date_objects}

# time.clock is deprecated and was removed in Python 3.8
real_clock = getattr(time, 'clock', None)

freeze_factories = []
tz_offsets = []
ignore_lists = []
tick_flags = []

try:
    # noinspection PyUnresolvedReferences
    real_uuid_generate_time = uuid._uuid_generate_time
    uuid_generate_time_attr = '_uuid_generate_time'
except AttributeError:
    # noinspection PyUnresolvedReferences
    uuid._load_system_functions()
    # noinspection PyUnresolvedReferences
    real_uuid_generate_time = uuid._generate_time_safe
    uuid_generate_time_attr = '_generate_time_safe'
except ImportError:
    real_uuid_generate_time = None
    uuid_generate_time_attr = None

try:
    # noinspection PyUnresolvedReferences
    real_uuid_create = uuid._UuidCreate
except (AttributeError, ImportError):
    real_uuid_create = None


# keep a cache of module attributes otherwise freezegun will need to analyze too many modules all the time
_GLOBAL_MODULES_CACHE = {}


def _get_module_attributes(module):
    result = []
    try:
        module_attributes = dir(module)
    except (ImportError, TypeError):
        return result
    for attribute_name in module_attributes:
        try:
            attribute_value = getattr(module, attribute_name)
        except (ImportError, AttributeError, TypeError):
            # For certain libraries, this can result in ImportError(_winreg) or AttributeError (celery)
            continue
        else:
            result.append((attribute_name, attribute_value))
    return result


def _setup_module_cache(module):
    date_attrs = []
    all_module_attributes = _get_module_attributes(module)
    for attribute_name, attribute_value in all_module_attributes:
        if id(attribute_value) in _real_time_object_ids:
            date_attrs.append((attribute_name, attribute_value))
    _GLOBAL_MODULES_CACHE[module.__name__] = (_get_module_attributes_hash(module), date_attrs)


def _get_module_attributes_hash(module):
    try:
        module_dir = dir(module)
    except (ImportError, TypeError):
        module_dir = []
    return f'{id(module)}-{hash(frozenset(module_dir))}'


def _get_cached_module_attributes(module):
    module_hash, cached_attrs = _GLOBAL_MODULES_CACHE.get(module.__name__, ('0', []))
    if _get_module_attributes_hash(module) == module_hash:
        return cached_attrs

    # cache miss: update the cache and return the refreshed value
    _setup_module_cache(module)
    # return the newly cached value
    module_hash, cached_attrs = _GLOBAL_MODULES_CACHE[module.__name__]
    return cached_attrs


_is_cpython = (
    hasattr(platform, 'python_implementation') and
    platform.python_implementation().lower() == "cpython"
)


call_stack_inspection_limit = 5


def _should_use_real_time():
    if not call_stack_inspection_limit:
        return False

    # Means stop() has already been called, so we can now return the real time
    if not ignore_lists:
        return True

    if not ignore_lists[-1]:
        return False

    frame = inspect.currentframe().f_back.f_back

    for _ in range(call_stack_inspection_limit):
        module_name = frame.f_globals.get('__name__')
        if module_name and module_name.startswith(ignore_lists[-1]):
            return True

        frame = frame.f_back
        if frame is None:
            break

    return False


def get_current_time():
    return freeze_factories[-1]()


def fake_time():
    if _should_use_real_time():
        return real_time()
    current_time = get_current_time()
    return calendar.timegm(current_time.timetuple()) + current_time.microsecond / 1000000.0

if _TIME_NS_PRESENT:
    def fake_time_ns():
        if _should_use_real_time():
            return real_time_ns()
        return int(int(fake_time()) * 1e9)


def fake_localtime(t=None):
    if t is not None:
        return real_localtime(t)
    if _should_use_real_time():
        return real_localtime()
    shifted_time = get_current_time() - datetime.timedelta(seconds=time.timezone)
    return shifted_time.timetuple()


def fake_gmtime(t=None):
    if t is not None:
        return real_gmtime(t)
    if _should_use_real_time():
        return real_gmtime()
    return get_current_time().timetuple()


def _get_fake_monotonic():
    # For monotonic timers like .monotonic(), .perf_counter(), etc
    current_time = get_current_time()
    return (
        calendar.timegm(current_time.timetuple()) +
        current_time.microsecond / 1e6
    )


def _get_fake_monotonic_ns():
    # For monotonic timers like .monotonic(), .perf_counter(), etc
    current_time = get_current_time()
    return (
        calendar.timegm(current_time.timetuple()) * 1000000 +
        current_time.microsecond
    ) * 1000


def fake_monotonic():
    if _should_use_real_time():
        return real_monotonic()

    return _get_fake_monotonic()


def fake_perf_counter():
    if _should_use_real_time():
        return real_perf_counter()

    return _get_fake_monotonic()


if _MONOTONIC_NS_PRESENT:
    def fake_monotonic_ns():
        if _should_use_real_time():
            return real_monotonic_ns()

        return _get_fake_monotonic_ns()


if _PERF_COUNTER_NS_PRESENT:
    def fake_perf_counter_ns():
        if _should_use_real_time():
            return real_perf_counter_ns()
        return _get_fake_monotonic_ns()


def fake_strftime(format, time_to_format=None):
    if time_to_format is None:
        if not _should_use_real_time():
            time_to_format = fake_localtime()

    if time_to_format is None:
        return real_strftime(format)
    else:
        return real_strftime(format, time_to_format)

if real_clock is not None:
    def fake_clock():
        if _should_use_real_time():
            return real_clock()

        if len(freeze_factories) == 1:
            return 0.0 if not tick_flags[-1] else real_clock()

        first_frozen_time = freeze_factories[0]()
        last_frozen_time = get_current_time()

        timedelta = (last_frozen_time - first_frozen_time)
        total_seconds = timedelta.total_seconds()

        if tick_flags[-1]:
            total_seconds += real_clock()

        return total_seconds


class FakeDateMeta(type):
    @classmethod
    def __instancecheck__(self, obj):
        return isinstance(obj, real_date)

    @classmethod
    def __subclasscheck__(cls, subclass):
        return issubclass(subclass, real_date)


def datetime_to_fakedatetime(datetime):
    return FakeDatetime(datetime.year,
                        datetime.month,
                        datetime.day,
                        datetime.hour,
                        datetime.minute,
                        datetime.second,
                        datetime.microsecond,
                        datetime.tzinfo)


def date_to_fakedate(date):
    return FakeDate(date.year,
                    date.month,
                    date.day)


class FakeDate(real_date, metaclass=FakeDateMeta):
    def __add__(self, other):
        result = real_date.__add__(self, other)
        if result is NotImplemented:
            return result
        return date_to_fakedate(result)

    def __sub__(self, other):
        result = real_date.__sub__(self, other)
        if result is NotImplemented:
            return result
        if isinstance(result, real_date):
            return date_to_fakedate(result)
        else:
            return result

    @classmethod
    def today(cls):
        result = cls._date_to_freeze() + cls._tz_offset()
        return date_to_fakedate(result)

    @staticmethod
    def _date_to_freeze():
        return get_current_time()

    @classmethod
    def _tz_offset(cls):
        return tz_offsets[-1]

FakeDate.min = date_to_fakedate(real_date.min)
FakeDate.max = date_to_fakedate(real_date.max)


class FakeDatetimeMeta(FakeDateMeta):
    @classmethod
    def __instancecheck__(self, obj):
        return isinstance(obj, real_datetime)

    @classmethod
    def __subclasscheck__(cls, subclass):
        return issubclass(subclass, real_datetime)


class FakeDatetime(real_datetime, FakeDate, metaclass=FakeDatetimeMeta):
    def __add__(self, other):
        result = real_datetime.__add__(self, other)
        if result is NotImplemented:
            return result
        return datetime_to_fakedatetime(result)

    def __sub__(self, other):
        result = real_datetime.__sub__(self, other)
        if result is NotImplemented:
            return result
        if isinstance(result, real_datetime):
            return datetime_to_fakedatetime(result)
        else:
            return result

    def astimezone(self, tz=None):
        if tz is None:
            tz = tzlocal()
        return datetime_to_fakedatetime(real_datetime.astimezone(self, tz))

    @classmethod
    def fromtimestamp(cls, t, tz=None):
        if tz is None:
            return real_datetime.fromtimestamp(
                    t, tz=dateutil.tz.tzoffset("freezegun", cls._tz_offset())
                ).replace(tzinfo=None)
        return datetime_to_fakedatetime(real_datetime.fromtimestamp(t, tz))

    def timestamp(self):
        if self.tzinfo is None:
            return (self - _EPOCH - self._tz_offset()).total_seconds()
        return (self - _EPOCHTZ).total_seconds()

    @classmethod
    def now(cls, tz=None):
        now = cls._time_to_freeze() or real_datetime.now()
        if tz:
            result = tz.fromutc(now.replace(tzinfo=tz)) + cls._tz_offset()
        else:
            result = now + cls._tz_offset()
        return datetime_to_fakedatetime(result)

    def date(self):
        return date_to_fakedate(self)

    @property
    def nanosecond(self):
        try:
            # noinspection PyUnresolvedReferences
            return real_datetime.nanosecond
        except AttributeError:
            return 0

    @classmethod
    def today(cls):
        return cls.now(tz=None)

    @classmethod
    def utcnow(cls):
        result = cls._time_to_freeze() or real_datetime.utcnow()
        return datetime_to_fakedatetime(result)

    @staticmethod
    def _time_to_freeze():
        if freeze_factories:
            return get_current_time()

    @classmethod
    def _tz_offset(cls):
        return tz_offsets[-1]


FakeDatetime.min = datetime_to_fakedatetime(real_datetime.min)
FakeDatetime.max = datetime_to_fakedatetime(real_datetime.max)


def convert_to_timezone_naive(time_to_freeze):
    """
    Converts a potentially timezone-aware datetime to be a naive UTC datetime
    """
    if time_to_freeze.tzinfo:
        time_to_freeze -= time_to_freeze.utcoffset()
        time_to_freeze = time_to_freeze.replace(tzinfo=None)
    return time_to_freeze


def pickle_fake_date(datetime_):
    # A pickle function for FakeDate
    return FakeDate, (
        datetime_.year,
        datetime_.month,
        datetime_.day,
    )


def pickle_fake_datetime(datetime_):
    # A pickle function for FakeDatetime
    return FakeDatetime, (
        datetime_.year,
        datetime_.month,
        datetime_.day,
        datetime_.hour,
        datetime_.minute,
        datetime_.second,
        datetime_.microsecond,
        datetime_.tzinfo,
    )


def _parse_time_to_freeze(time_to_freeze_str):
    """Parses all the possible inputs for freeze_time
    :returns: a naive ``datetime.datetime`` object
    """
    if time_to_freeze_str is None:
        time_to_freeze_str = datetime.datetime.utcnow()

    if isinstance(time_to_freeze_str, datetime.datetime):
        time_to_freeze = time_to_freeze_str
    elif isinstance(time_to_freeze_str, datetime.date):
        time_to_freeze = datetime.datetime.combine(time_to_freeze_str, datetime.time())
    elif isinstance(time_to_freeze_str, datetime.timedelta):
        time_to_freeze = datetime.datetime.utcnow() + time_to_freeze_str
    else:
        time_to_freeze = parser.parse(time_to_freeze_str)

    return convert_to_timezone_naive(time_to_freeze)


def _parse_tz_offset(tz_offset):
    if isinstance(tz_offset, datetime.timedelta):
        return tz_offset
    else:
        return datetime.timedelta(hours=tz_offset)


class TickingDateTimeFactory:

    def __init__(self, time_to_freeze, start):
        self.time_to_freeze = time_to_freeze
        self.start = start

    def __call__(self):
        return self.time_to_freeze + (real_datetime.now() - self.start)


class FrozenDateTimeFactory:

    def __init__(self, time_to_freeze):
        self.time_to_freeze = time_to_freeze

    def __call__(self):
        return self.time_to_freeze

    def tick(self, delta=datetime.timedelta(seconds=1)):
        if isinstance(delta, numbers.Real):
            # noinspection PyTypeChecker
            self.time_to_freeze += datetime.timedelta(seconds=delta)
        else:
            self.time_to_freeze += delta

    def move_to(self, target_datetime):
        """Moves frozen date to the given ``target_datetime``"""
        target_datetime = _parse_time_to_freeze(target_datetime)
        delta = target_datetime - self.time_to_freeze
        self.tick(delta=delta)


class StepTickTimeFactory:

    def __init__(self, time_to_freeze, step_width):
        self.time_to_freeze = time_to_freeze
        self.step_width = step_width

    def __call__(self):
        return_time = self.time_to_freeze
        self.tick()
        return return_time

    def tick(self, delta=None):
        if not delta:
            delta = datetime.timedelta(seconds=self.step_width)
        self.time_to_freeze += delta

    def update_step_width(self, step_width):
        self.step_width = step_width

    def move_to(self, target_datetime):
        """Moves frozen date to the given ``target_datetime``"""
        target_datetime = _parse_time_to_freeze(target_datetime)
        delta = target_datetime - self.time_to_freeze
        self.tick(delta=delta)


class _freeze_time:

    def __init__(self, time_to_freeze_str, tz_offset, ignore, tick, as_arg, as_kwarg, auto_tick_seconds):
        self.time_to_freeze = _parse_time_to_freeze(time_to_freeze_str)
        self.tz_offset = _parse_tz_offset(tz_offset)
        self.ignore = tuple(ignore)
        self.tick = tick
        self.auto_tick_seconds = auto_tick_seconds
        self.undo_changes = []
        self.modules_at_start = set()
        self.as_arg = as_arg
        self.as_kwarg = as_kwarg

    def __call__(self, func):
        if inspect.isclass(func):
            return self.decorate_class(func)
        elif inspect.iscoroutinefunction(func):
            return self.decorate_coroutine(func)
        return self.decorate_callable(func)

    def decorate_class(self, klass):
        if issubclass(klass, unittest.TestCase):
            # If it's a TestCase, we freeze time around setup and teardown, as well
            # as for every test case. This requires some care to avoid freezing
            # the time pytest sees, as otherwise this would distort the reported
            # timings.

            orig_setUpClass = klass.setUpClass
            orig_tearDownClass = klass.tearDownClass

            # noinspection PyDecorator
            @classmethod
            def setUpClass(cls):
                self.start()
                if orig_setUpClass is not None:
                    orig_setUpClass()
                self.stop()

            # noinspection PyDecorator
            @classmethod
            def tearDownClass(cls):
                self.start()
                if orig_tearDownClass is not None:
                    orig_tearDownClass()
                self.stop()

            klass.setUpClass = setUpClass
            klass.tearDownClass = tearDownClass

            orig_setUp = klass.setUp
            orig_tearDown = klass.tearDown

            def setUp(*args, **kwargs):
                self.start()
                if orig_setUp is not None:
                    orig_setUp(*args, **kwargs)

            def tearDown(*args, **kwargs):
                if orig_tearDown is not None:
                    orig_tearDown(*args, **kwargs)
                self.stop()

            klass.setUp = setUp
            klass.tearDown = tearDown

            return klass

        else:

            seen = set()

            klasses = klass.mro()
            for base_klass in klasses:
                for (attr, attr_value) in base_klass.__dict__.items():
                    if attr.startswith('_') or attr in seen:
                        continue
                    seen.add(attr)

                    if not callable(attr_value) or inspect.isclass(attr_value):
                        continue

                    try:
                        setattr(klass, attr, self(attr_value))
                    except (AttributeError, TypeError):
                        # Sometimes we can't set this for built-in types and custom callables
                        continue
            return klass

    def __enter__(self):
        return self.start()

    def __exit__(self, *args):
        self.stop()

    def start(self):

        if self.auto_tick_seconds:
            freeze_factory = StepTickTimeFactory(self.time_to_freeze, self.auto_tick_seconds)
        elif self.tick:
            freeze_factory = TickingDateTimeFactory(self.time_to_freeze, real_datetime.now())
        else:
            freeze_factory = FrozenDateTimeFactory(self.time_to_freeze)

        is_already_started = len(freeze_factories) > 0
        freeze_factories.append(freeze_factory)
        tz_offsets.append(self.tz_offset)
        ignore_lists.append(self.ignore)
        tick_flags.append(self.tick)

        if is_already_started:
            return freeze_factory

        # Change the modules
        datetime.datetime = FakeDatetime
        datetime.date = FakeDate

        time.time = fake_time
        time.monotonic = fake_monotonic
        time.perf_counter = fake_perf_counter
        time.localtime = fake_localtime
        time.gmtime = fake_gmtime
        time.strftime = fake_strftime
        if uuid_generate_time_attr:
            setattr(uuid, uuid_generate_time_attr, None)
        uuid._UuidCreate = None
        uuid._last_timestamp = None

        copyreg.dispatch_table[real_datetime] = pickle_fake_datetime
        copyreg.dispatch_table[real_date] = pickle_fake_date

        # Change any place where the module had already been imported
        to_patch = [
            ('real_date', real_date, FakeDate),
            ('real_datetime', real_datetime, FakeDatetime),
            ('real_gmtime', real_gmtime, fake_gmtime),
            ('real_localtime', real_localtime, fake_localtime),
            ('real_monotonic', real_monotonic, fake_monotonic),
            ('real_perf_counter', real_perf_counter, fake_perf_counter),
            ('real_strftime', real_strftime, fake_strftime),
            ('real_time', real_time, fake_time),
        ]

        if _TIME_NS_PRESENT:
            time.time_ns = fake_time_ns
            to_patch.append(('real_time_ns', real_time_ns, fake_time_ns))

        if _MONOTONIC_NS_PRESENT:
            time.monotonic_ns = fake_monotonic_ns
            to_patch.append(('real_monotonic_ns', real_monotonic_ns, fake_monotonic_ns))

        if _PERF_COUNTER_NS_PRESENT:
            time.perf_counter_ns = fake_perf_counter_ns
            to_patch.append(('real_perf_counter_ns', real_perf_counter_ns, fake_perf_counter_ns))

        if real_clock is not None:
            # time.clock is deprecated and was removed in Python 3.8
            time.clock = fake_clock
            to_patch.append(('real_clock', real_clock, fake_clock))

        self.fake_names = tuple(fake.__name__ for real_name, real, fake in to_patch)
        self.reals = {id(fake): real for real_name, real, fake in to_patch}
        fakes = {id(real): fake for real_name, real, fake in to_patch}
        add_change = self.undo_changes.append

        # Save the current loaded modules
        self.modules_at_start = set(sys.modules.keys())

        with warnings.catch_warnings():
            warnings.filterwarnings('ignore')

            for mod_name, module in list(sys.modules.items()):
                if mod_name is None or module is None or mod_name == __name__:
                    continue
                elif mod_name.startswith(self.ignore) or mod_name.endswith('.six.moves'):
                    continue
                elif (not hasattr(module, "__name__") or module.__name__ in ('datetime', 'time')):
                    continue

                module_attrs = _get_cached_module_attributes(module)
                for attribute_name, attribute_value in module_attrs:
                    fake = fakes.get(id(attribute_value))
                    if fake:
                        setattr(module, attribute_name, fake)
                        add_change((module, attribute_name, attribute_value))

        return freeze_factory

    def stop(self):
        freeze_factories.pop()
        ignore_lists.pop()
        tick_flags.pop()
        tz_offsets.pop()

        if not freeze_factories:
            datetime.datetime = real_datetime
            datetime.date = real_date
            copyreg.dispatch_table.pop(real_datetime)
            copyreg.dispatch_table.pop(real_date)
            for module, module_attribute, original_value in self.undo_changes:
                setattr(module, module_attribute, original_value)
            self.undo_changes = []

            # Restore modules loaded after start()
            modules_to_restore = set(sys.modules.keys()) - self.modules_at_start
            self.modules_at_start = set()
            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                for mod_name in modules_to_restore:
                    module = sys.modules.get(mod_name, None)
                    if mod_name is None or module is None:
                        continue
                    elif mod_name.startswith(self.ignore) or mod_name.endswith('.six.moves'):
                        continue
                    elif not hasattr(module, "__name__") or module.__name__ in ('datetime', 'time'):
                        continue
                    for module_attribute in dir(module):

                        if module_attribute in self.fake_names:
                            continue
                        try:
                            attribute_value = getattr(module, module_attribute)
                        except (ImportError, AttributeError, TypeError):
                            # For certain libraries, this can result in ImportError(_winreg) or AttributeError (celery)
                            continue

                        real = self.reals.get(id(attribute_value))
                        if real:
                            setattr(module, module_attribute, real)

            time.time = real_time
            time.monotonic = real_monotonic
            time.perf_counter = real_perf_counter
            time.gmtime = real_gmtime
            time.localtime = real_localtime
            time.strftime = real_strftime
            time.clock = real_clock

            if _TIME_NS_PRESENT:
                time.time_ns = real_time_ns

            if _MONOTONIC_NS_PRESENT:
                time.monotonic_ns = real_monotonic_ns

            if _PERF_COUNTER_NS_PRESENT:
                time.perf_counter_ns = real_perf_counter_ns

            if uuid_generate_time_attr:
                setattr(uuid, uuid_generate_time_attr, real_uuid_generate_time)
            uuid._UuidCreate = real_uuid_create
            uuid._last_timestamp = None

    def decorate_coroutine(self, coroutine):
        return wrap_coroutine(self, coroutine)

    def decorate_callable(self, func):
        def wrapper(*args, **kwargs):
            with self as time_factory:
                if self.as_arg and self.as_kwarg:
                    assert False, "You can't specify both as_arg and as_kwarg at the same time. Pick one."
                elif self.as_arg:
                    result = func(time_factory, *args, **kwargs)
                elif self.as_kwarg:
                    kwargs[self.as_kwarg] = time_factory
                    result = func(*args, **kwargs)
                else:
                    result = func(*args, **kwargs)
            return result
        functools.update_wrapper(wrapper, func)

        return wrapper


def freeze_time(time_to_freeze=None, tz_offset=0, ignore=None, tick=False, as_arg=False, as_kwarg='',
                auto_tick_seconds=0):
    acceptable_times = (type(None), str, datetime.date, datetime.timedelta,
             types.FunctionType, types.GeneratorType)

    if MayaDT is not None:
        acceptable_times += MayaDT,

    if not isinstance(time_to_freeze, acceptable_times):
        raise TypeError(('freeze_time() expected None, a string, date instance, datetime '
                         'instance, MayaDT, timedelta instance, function or a generator, but got '
                         'type {}.').format(type(time_to_freeze)))
    if tick and not _is_cpython:
        raise SystemError('Calling freeze_time with tick=True is only compatible with CPython')

    if isinstance(time_to_freeze, types.FunctionType):
        return freeze_time(time_to_freeze(), tz_offset, ignore, tick, as_arg, as_kwarg, auto_tick_seconds)

    if isinstance(time_to_freeze, types.GeneratorType):
        return freeze_time(next(time_to_freeze), tz_offset, ignore, tick, as_arg, as_kwarg, auto_tick_seconds)

    if MayaDT is not None and isinstance(time_to_freeze, MayaDT):
        return freeze_time(time_to_freeze.datetime(), tz_offset, ignore,
                           tick, as_arg, as_kwarg, auto_tick_seconds)

    if ignore is None:
        ignore = []
    ignore = ignore[:]
    if config.settings.default_ignore_list:
        ignore.extend(config.settings.default_ignore_list)

    return _freeze_time(
        time_to_freeze_str=time_to_freeze,
        tz_offset=tz_offset,
        ignore=ignore,
        tick=tick,
        as_arg=as_arg,
        as_kwarg=as_kwarg,
        auto_tick_seconds=auto_tick_seconds,
    )


# Setup adapters for sqlite
try:
    # noinspection PyUnresolvedReferences
    import sqlite3
except ImportError:
    # Some systems have trouble with this
    pass
else:
    # These are copied from Python sqlite3.dbapi2
    def adapt_date(val):
        return val.isoformat()

    def adapt_datetime(val):
        return val.isoformat(" ")

    sqlite3.register_adapter(FakeDate, adapt_date)
    sqlite3.register_adapter(FakeDatetime, adapt_datetime)


# Setup converters for pymysql
try:
    import pymysql.converters
except ImportError:
    pass
else:
    pymysql.converters.encoders[FakeDate] = pymysql.converters.encoders[real_date]
    pymysql.converters.conversions[FakeDate] = pymysql.converters.encoders[real_date]
    pymysql.converters.encoders[FakeDatetime] = pymysql.converters.encoders[real_datetime]
    pymysql.converters.conversions[FakeDatetime] = pymysql.converters.encoders[real_datetime]



# Copyright (c) 2015-2019, Activision Publishing, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import sys

if sys.version_info[0] == 3:
    str_types = (str,)
else:
    str_types = (basestring,)

__tracebackhide__ = True


def contents_of(file, encoding='utf-8'):
    """Helper to read the contents of the given file or path into a string with the given encoding.

    Args:
        file: a *path-like object* (aka a file name) or a *file-like object* (aka a file)
        encoding (str): the target encoding.  Defaults to ``utf-8``, other useful encodings are ``ascii`` and ``latin-1``.

    Examples:
        Usage::

            from assertpy import assert_that, contents_of

            contents = contents_of('foo.txt')
            assert_that(contents).starts_with('foo').ends_with('bar').contains('oob')

    Returns:
        str: returns the file contents as a string

    Raises:
        IOError: if file not found
        TypeError: if file is not a *path-like object* or a *file-like object*
    """
    try:
        contents = file.read()
    except AttributeError:
        try:
            with open(file, 'r') as fp:
                contents = fp.read()
        except TypeError:
            raise ValueError('val must be file or path, but was type <%s>' % type(file).__name__)
        except OSError:
            if not isinstance(file, str_types):
                raise ValueError('val must be file or path, but was type <%s>' % type(file).__name__)
            raise

    if sys.version_info[0] == 3 and type(contents) is bytes:
        # in PY3 force decoding of bytes to target encoding
        return contents.decode(encoding, 'replace')
    elif sys.version_info[0] == 2 and encoding == 'ascii':
        # in PY2 force encoding back to ascii
        return contents.encode('ascii', 'replace')
    else:
        # in all other cases, try to decode to target encoding
        try:
            return contents.decode(encoding, 'replace')
        except AttributeError:
            pass
    # if all else fails, just return the contents "as is"
    return contents


class FileMixin(object):
    """File assertions mixin."""

    def exists(self):
        """Asserts that val is a path and that it exists.

        Examples:
            Usage::

                assert_that('myfile.txt').exists()
                assert_that('mydir').exists()

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val does **not** exist
        """
        if not isinstance(self.val, str_types):
            raise TypeError('val is not a path')
        if not os.path.exists(self.val):
            self.error('Expected <%s> to exist, but was not found.' % self.val)
        return self

    def does_not_exist(self):
        """Asserts that val is a path and that it does *not* exist.

        Examples:
            Usage::

                assert_that('missing.txt').does_not_exist()
                assert_that('missing_dir').does_not_exist()

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val **does** exist
        """
        if not isinstance(self.val, str_types):
            raise TypeError('val is not a path')
        if os.path.exists(self.val):
            self.error('Expected <%s> to not exist, but was found.' % self.val)
        return self

    def is_file(self):
        """Asserts that val is a *file* and that it exists.

        Examples:
            Usage::

                assert_that('myfile.txt').is_file()

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val does **not** exist, or is **not** a file
        """
        self.exists()
        if not os.path.isfile(self.val):
            self.error('Expected <%s> to be a file, but was not.' % self.val)
        return self

    def is_directory(self):
        """Asserts that val is a *directory* and that it exists.

        Examples:
            Usage::

                assert_that('mydir').is_directory()

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val does **not** exist, or is **not** a directory
        """
        self.exists()
        if not os.path.isdir(self.val):
            self.error('Expected <%s> to be a directory, but was not.' % self.val)
        return self

    def is_named(self, filename):
        """Asserts that val is an existing path to a file and that file is named filename.

        Args:
            filename: the expected filename

        Examples:
            Usage::

                assert_that('/path/to/mydir/myfile.txt').is_named('myfile.txt')

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val does **not** exist, or is **not** a file, or is **not** named the given filename
        """
        self.is_file()
        if not isinstance(filename, str_types):
            raise TypeError('given filename arg must be a path')
        val_filename = os.path.basename(os.path.abspath(self.val))
        if val_filename != filename:
            self.error('Expected filename <%s> to be equal to <%s>, but was not.' % (val_filename, filename))
        return self

    def is_child_of(self, parent):
        """Asserts that val is an existing path to a file and that file is a child of parent.

        Args:
            parent: the expected parent directory

        Examples:
            Usage::

                assert_that('/path/to/mydir/myfile.txt').is_child_of('mydir')
                assert_that('/path/to/mydir/myfile.txt').is_child_of('to')
                assert_that('/path/to/mydir/myfile.txt').is_child_of('path')

        Returns:
            AssertionBuilder: returns this instance to chain to the next assertion

        Raises:
            AssertionError: if val does **not** exist, or is **not** a file, or is **not** a child of the given directory
        """
        self.is_file()
        if not isinstance(parent, str_types):
            raise TypeError('given parent directory arg must be a path')
        val_abspath = os.path.abspath(self.val)
        parent_abspath = os.path.abspath(parent)
        if not val_abspath.startswith(parent_abspath):
            self.error('Expected file <%s> to be a child of <%s>, but was not.' % (val_abspath, parent_abspath))
        return self



import functools
from typing import Any, Callable, TypeVar, cast


_CallableT = TypeVar("_CallableT", bound=Callable[..., Any])


def wrap_coroutine(api: Any, coroutine: _CallableT) -> _CallableT:
    @functools.wraps(coroutine)
    async def wrapper(*args: Any, **kwargs: Any) -> Any:
        with api as time_factory:
            if api.as_arg:
                result = await coroutine(time_factory, *args, **kwargs)
            else:
                result = await coroutine(*args, **kwargs)
        return result

    return cast(_CallableT, wrapper)



from __future__ import absolute_import
from .assertpy import assert_that, assert_warn, soft_assertions, fail, soft_fail, add_extension, remove_extension, WarningLoggingAdapter, __version__
from .file import contents_of



import logging
import os
import subprocess
from abc import ABC
from datetime import datetime
from pathlib import Path
from typing import Dict

from jsonschema.validators import validate

from bashckup.actuators import writers
from bashckup.actuators.actuators import PythonActuator, ActuatorMetadata
from bashckup.actuators.exceptions import RunningException, ParameterException


class AbstractPostBackup(PythonActuator, ABC):

    def _actuator_type(self) -> str:
        return 'post-backup'


class CleanFolderPostBackup(AbstractPostBackup):
    validation_schema = {'type': 'object', 'properties': {
        'retention': {
            'type': 'integer',
            'minimum': 1,
            'description': 'Retention duration in days'}},
                         'required': ['retention'],
                         'additionalProperties': False}

    def __init__(self, global_context: dict, args: dict, metadata: Dict[str, Dict[str, ActuatorMetadata]] = None):
        super().__init__(global_context, args, metadata)
        self.retention = None
        self._output_directory = None

    @staticmethod
    def module_name() -> str:
        return 'cleanFolder'

    def _get_params(self) -> None:
        validate(self._args, self.validation_schema)

        self.retention = self._args['retention']

    def _prepare_run_backup(self) -> dict:
        # Validate metadata
        output_directories = [v.get('output-directory') for (i, v) in self._metadata['writer'].items()]
        if len(output_directories) != 1:  # Because we need at least one, and it can not be greater than 1
            raise ValueError('output-directory must be defined in writer module')
        self._output_directory = output_directories[0]

        file_preservation_window = max(
            [v.get('file-preservation-window') for (i, v) in self._metadata['reader'].items()])
        if file_preservation_window is not None and self.retention < file_preservation_window:
            logging.warning(
                f'WARNING: retention [{self.retention}] is lower than the minimum possible '
                f'[{file_preservation_window}]. Retention value is override by the minimum')
            self.retention = file_preservation_window

        files_to_remove = []
        today = datetime.today()
        with os.scandir(self._output_directory) as it:
            entry: os.DirEntry
            for entry in it:
                matches = writers.datetimeOutputFileRegex.search(entry.name)
                if matches is None or len(matches.groups()) != 1:
                    continue
                creation_date = datetime.fromisoformat(matches.group(1))
                diff_days = (today - creation_date).days
                if diff_days >= self.retention:
                    files_to_remove.append(entry.path)
        return {'files-to-remove': files_to_remove}

    def _run_backup(self, args: dict) -> None:
        for file_path in args['files-to-remove']:
            try:
                os.remove(file_path)
                logging.info('File [%s] removed', file_path)
            except OSError as e:
                raise RunningException(f'Unable to remove file [{file_path}].\nReason: {e}') from e

    def _dry_run_backup(self, args: dict) -> None:
        for file_path in args['files-to-remove']:
            logging.info(f'File [{file_path}] would have been deleted.')

    def _prepare_run_restore(self) -> dict:
        return {}

    def _run_restore(self, args: dict) -> None:
        pass

    def _dry_run_restore(self, args: dict) -> None:
        logging.debug(f'Module [{self.module_name()}] have nothing to do')


class RsyncPostBackup(AbstractPostBackup):
    validation_schema = {'type': 'object', 'properties': {
        'ip-addr': {
            'type': 'string',
            'description': 'IP address of remote host'},
        'dest-module': {
            'type': 'string',
            'description': 'Destination rsyncd module'},
        'dest-folder': {
            'type': 'string',
            'description': 'Destination folder'},
        'user': {
            'type': 'string',
            'description': 'Username to use to log in'},
        'password-file': {
            'type': 'string',
            'description': 'Path to the file that contains the password'}},
                         'required': ['ip-addr', 'dest-folder', 'user'],
                         'additionalProperties': False}

    def __init__(self, global_context: dict, args: dict, metadata: Dict[str, Dict[str, ActuatorMetadata]] = None):
        super().__init__(global_context, args, metadata)
        self._output_directory = None

    @staticmethod
    def module_name() -> str:
        return 'rsync'

    def _get_params(self) -> None:
        validate(self._args, self.validation_schema)

        self.ip_addr = self._args['ip-addr']
        self.dest_module = self._args.get('dest-module')
        self.dest_folder = self._args['dest-folder']
        self.user = self._args['user']
        self.password_file = self._args.get('password-file')
        if self.password_file is not None:
            password_file = Path(self._args['password-file'])
            if not password_file.is_file():
                raise ParameterException('password-file [' + self._args['password-file'] + '] doesn\'t exists',
                                         'password-file', self._backup_id, self.module_name())
            if not os.access(password_file, os.R_OK):
                raise ParameterException('password-file [' + self._args['password-file'] + '] is not readable',
                                         'password-file', self._backup_id, self.module_name())
            password_file_stat = os.stat(password_file)
            if password_file_stat.st_mode & 0o077 != 0o0:
                raise ParameterException(
                    'File [' + self._args['password-file'] + '] must not be readable from group and from others',
                    'password-file', self._backup_id,
                    self.module_name())
            if password_file_stat.st_uid != os.getuid():
                raise ParameterException(
                    'File [' + self._args['password-file'] + '] is not owned by current user',
                    'password-file', self._backup_id,
                    self.module_name())

    def _prepare_run_backup(self) -> dict:
        # Validate metadata
        self._validate_register_metadata()

        cmd = ['rsync']
        if self._verbose:
            cmd.append('--progress')
        cmd.extend('--archive --no-inc-recursive --exclude={"lost+found/"} --delete-after'.split(' '))
        if self.password_file is not None:
            cmd.extend(['--password-file', self.password_file])
        cmd.append(self._output_directory)
        destination = self.user + '@' + self.ip_addr
        if self.dest_module is not None:
            destination += '::' + self.dest_module + '/'
        else:
            destination += ':/'
        if self.dest_folder[0] == '/':  # Remove first slash if exists
            destination += self.dest_folder[1:]
        else:
            destination += self.dest_folder
        cmd.append(destination)

        return {'cmd': cmd}

    def _run_backup(self, args: dict) -> None:
        process = subprocess.run(args['cmd'], capture_output=True, shell=False, text=True)
        if process.returncode != 0:
            raise RunningException('Error during execution of rsync\n'
                                   f'Rsync output: {process.stderr}\n'
                                   f'''Command executed: {' '.join(process.args)}''')

    def _dry_run_backup(self, args: dict) -> None:
        logging.info(f'''Command [{' '.join(args['cmd'])}] would have been ran.''')

    def _prepare_run_restore(self) -> dict:
        # Validate metadata
        self._validate_register_metadata()

        cmd = ['rsync']
        if self._verbose:
            cmd.append('--progress')
        cmd.extend('--archive --no-inc-recursive --exclude={"lost+found/"} --delete-after'.split(' '))
        if self.password_file is not None:
            cmd.extend(['--password-file', self.password_file])
        remote_src = self.user + '@' + self.ip_addr
        if self.dest_module is not None:
            remote_src += '::' + self.dest_module + '/'
        else:
            remote_src += ':/'
        if self.dest_folder[0] == '/':  # Remove first slash if exists
            remote_src += self.dest_folder[1:]
        else:
            remote_src += self.dest_folder
        cmd.append(remote_src)
        cmd.append(self._output_directory)

        return {'cmd': cmd}

    def _validate_register_metadata(self):
        """
        Validates metadata and register attributes with it
        """
        output_directories = [v.get('output-directory') for (i, v) in self._metadata['writer'].items()]
        if len(output_directories) != 1:  # Because we need at least one, and it can not be greater than 1
            raise ValueError('output-directory must be defined in writer module')
        self._output_directory = output_directories[0]
        # Add trailing slash to rsync the content of the folder and not the folder
        if self._output_directory[-1] != '/':
            self._output_directory = self._output_directory + '/'

    def _run_restore(self, args: dict) -> None:
        process = subprocess.run(args['cmd'], capture_output=True, shell=False, text=True)
        if process.returncode != 0:
            raise RunningException('Error during execution of rsync\n'
                                   f'Rsync output: {process.stderr}\n'
                                   f'''Command executed: {' '.join(process.args)}''')

    def _dry_run_restore(self, args: dict) -> None:
        logging.info(f'''Command [{' '.join(args['cmd'])}] would have been ran.''')
